"""
The core module of the PaYnter project, this class will do most of the work when you use this library.
"""

#External modules
import numpy as N
import PIL.Image
import PIL.ImageOps
import PIL.ImageFont
import PIL.ImageDraw 


#PaYnter Modules
from .blendModes import *
from .brush import *
from .utils import *
from .image import *
from .layer import *
import paynter.config as config


from numba import jit, float32
import time


'''
TODO:
-add asserts to main functions
'''

######################################################################
# Paynter class
######################################################################
#The paynter class is the object that draw stuff on a layer using a brush and a color
class Paynter:
	"""
    This class is the main object of the library and the one that will draw everything you ask.
	To create this class you can use the default constructor.
       
    .. code-block:: python

    	from paynter import *
    	P = Paynter()
	
    """
	brush = 0
	layer = 0
	color = Color(0, 0, 0, 1)
	secondColor = Color(1,1,1,1)
	image = 0
	mirrorMode = 0

	#Init the paynter
	def __init__(self):
		#Setup some stuff
		config.CANVAS_SIZE = int(config.REAL_CANVAS_SIZE/config.DOWNSAMPLING)
		self.image = Image()



	######################################################################
	# Level 0 Functions, needs downsampling
	######################################################################
	#Draw a line between two points
	def drawLine(self, x1, y1, x2, y2, silent=False):
		"""
		Draws a line on the current :py:class:`Layer` with the current :py:class:`Brush`.
		Coordinates are relative to the original layer size WITHOUT downsampling applied.

		:param x1: Starting X coordinate.
		:param y1: Starting Y coordinate.
		:param x2: End X coordinate.
		:param y2: End Y coordinate.
		:rtype: Nothing.
		"""
		start = time.time()

		#Downsample the coordinates
		x1 = int(x1/config.DOWNSAMPLING)
		x2 = int(x2/config.DOWNSAMPLING)
		y1 = int(y1/config.DOWNSAMPLING)
		y2 = int(y2/config.DOWNSAMPLING)
		if not silent :
			print('drawing line from: '+str((x1,y1))+' to: '+str((x2,y2)))

		#Calculate the direction and the length of the step
		direction = N.arctan2(y2 - y1, x2 - x1)
		length = self.brush.spacing

		#Prepare the loop
		x, y = x1, y1
		totalSteps = int(N.sqrt((x2 - x)**2 + (y2 - y)**2)/length)
		lay = self.image.getActiveLayer()
		col = self.color
		secCol = self.secondColor
		mirr = self.mirrorMode

		#If I use source caching..
		if self.brush.usesSourceCaching:
			#..than optimize it for faster drawing
			laydata = lay.data
			x -= self.brush.brushSize*0.5
			y -= self.brush.brushSize*0.5
			colbrsource = self.brush.coloredBrushSource
			canvSize = config.CANVAS_SIZE
			brmask = self.brush.brushMask
			for _ in range(totalSteps):
				#Make the dab on this point
				applyMirroredDab_jit(mirr, laydata, int(x), int(y), colbrsource.copy(), canvSize, brmask)

				#Mode the point for the next step and update the distances
				x += lendir_x(length, direction)
				y += lendir_y(length, direction)
		#..if I don't use source caching..
		else:
			#..do the normal drawing
			for _ in range(totalSteps):
				#Make the dab on this point
				self.brush.makeDab(lay, int(x), int(y), col, secCol, mirror=mirr)

				#Mode the point for the next step and update the distances
				x += lendir_x(length, direction)
				y += lendir_y(length, direction)
			
		
	#Draw a single dab
	def drawPoint(self, x, y, silent=True):
		"""
		Draws a point on the current :py:class:`Layer` with the current :py:class:`Brush`.
		Coordinates are relative to the original layer size WITHOUT downsampling applied.

		:param x1: Point X coordinate.
		:param y1: Point Y coordinate.
		:rtype: Nothing.
		"""
		start = time.time()

		#Downsample the coordinates
		x = int(x/config.DOWNSAMPLING)
		y = int(y/config.DOWNSAMPLING)

		#Apply the dab with or without source caching
		if self.brush.usesSourceCaching:
			applyMirroredDab_jit(self.mirrorMode, self.image.getActiveLayer().data, int(x-self.brush.brushSize*0.5), int(y-self.brush.brushSize*0.5), self.brush.coloredBrushSource.copy(), config.CANVAS_SIZE, self.brush.brushMask)
		else:
			self.brush.makeDab(self.image.getActiveLayer(), int(x), int(y), self.color, self.secondColor, mirror=self.mirrorMode)
		config.AVGTIME.append(time.time()-start)
		
	######################################################################
	# Level 1 Functions, calls Level 0 functions, no downsampling
	######################################################################
	#Draw a path from a series of points
	def drawPath(self, pointList):
		"""
		Draws a series of lines on the current :py:class:`Layer` with the current :py:class:`Brush`.
		No interpolation is applied to these point and :py:meth:`drawLine` will be used to connect all the points lineraly.
		Coordinates are relative to the original layer size WITHOUT downsampling applied.
		
		:param pointList: A list of point like :code:`[(0, 0), (100, 100), (100, 200)]`.
		:rtype: Nothing.
		"""
		self.drawLine(pointList[0][0], pointList[0][1], pointList[1][0], pointList[1][1])
		i = 1
		while i<len(pointList)-1:
			self.drawLine(pointList[i][0], pointList[i][1], pointList[i+1][0], pointList[i+1][1])
			i+=1

	#Draw a path from a series of points
	def drawClosedPath(self, pointList):
		"""
		Draws a closed series of lines on the current :py:class:`Layer` with the current :py:class:`Brush`.
		No interpolation is applied to these point and :py:meth:`drawLine` will be used to connect all the points lineraly.
		Coordinates are relative to the original layer size WITHOUT downsampling applied.
		
		:param pointList: A list of point like :code:`[(0, 0), (100, 100), (100, 200)]`.
		:rtype: Nothing.
		"""
		self.drawLine(pointList[0][0], pointList[0][1], pointList[1][0], pointList[1][1])
		i = 1
		while i<len(pointList)-1:
			self.drawLine(pointList[i][0], pointList[i][1], pointList[i+1][0], pointList[i+1][1])
			i+=1
		self.drawLine(pointList[-1][0], pointList[-1][1], pointList[0][0], pointList[0][1])
		
	#Draw a rectangle
	def drawRect(self, x1, y1, x2, y2, angle=0):
		"""
		Draws a rectangle on the current :py:class:`Layer` with the current :py:class:`Brush`.
		Coordinates are relative to the original layer size WITHOUT downsampling applied.
		
		:param x1: The X of the top-left corner of the rectangle.
		:param y1: The Y of the top-left corner of the rectangle.
		:param x2: The X of the bottom-right corner of the rectangle.
		:param y2: The Y of the bottom-right corner of the rectangle.
		:param angle: An angle (in degrees) of rotation around the center of the rectangle.
		:rtype: Nothing.
		"""
		vertices = [[x1,y1],[x2,y1],[x2,y2],[x1,y2],]
		rotatedVertices = rotateMatrix(vertices, (x1+x2)*0.5, (y1+y2)*0.5, angle)
		self.drawClosedPath(rotatedVertices)

	#Fill the current layer with a color
	def fillLayerWithColor(self, color):
		"""
		Fills the current :py:class:`Layer` with the current :py:class:`Color`.
		
		:param color: The :py:class:`Color` to apply to the layer.
		:rtype: Nothing.
		"""
		layer = self.image.getActiveLayer().data
		colorRGBA = color.get_0_255()
		layer[:,:,0] = colorRGBA[0]
		layer[:,:,1] = colorRGBA[1]
		layer[:,:,2] = colorRGBA[2]
		layer[:,:,3] = colorRGBA[3]

	#Add border to image
	def addBorder(self, width, color=None):
		"""
		Add a border to the current :py:class:`Layer`.
		
		:param width: The width of the border.
		:param color: The :py:class:`Color` of the border, current :py:class:`Color` is the default value.
		:rtype: Nothing.
		"""
		width = int(width/config.DOWNSAMPLING)
		if color==None:
			color = self.color
		layer = self.image.getActiveLayer().data
		colorRGBA = color.get_0_255()
		print('adding border'+str(colorRGBA)+str(width)+str(layer.shape))
		layer[0:width,:,0] = colorRGBA[0]
		layer[0:width,:,1] = colorRGBA[1]
		layer[0:width,:,2] = colorRGBA[2]
		layer[0:width,:,3] = colorRGBA[3]

		layer[:,0:width,0] = colorRGBA[0]
		layer[:,0:width,1] = colorRGBA[1]
		layer[:,0:width,2] = colorRGBA[2]
		layer[:,0:width,3] = colorRGBA[3]

		layer[layer.shape[0]-width:layer.shape[0],:,0] = colorRGBA[0]
		layer[layer.shape[0]-width:layer.shape[0],:,1] = colorRGBA[1]
		layer[layer.shape[0]-width:layer.shape[0],:,2] = colorRGBA[2]
		layer[layer.shape[0]-width:layer.shape[0],:,3] = colorRGBA[3]

		layer[:,layer.shape[1]-width:layer.shape[1],0] = colorRGBA[0]
		layer[:,layer.shape[1]-width:layer.shape[1],1] = colorRGBA[1]
		layer[:,layer.shape[1]-width:layer.shape[1],2] = colorRGBA[2]
		layer[:,layer.shape[1]-width:layer.shape[1],3] = colorRGBA[3]




	######################################################################
	# Setters, getters, and more
	######################################################################
	#Setter for color, takes 0-255 RGBA
	def setColor(self, color):
		"""
		Sets the current :py:class:`Color` to use.
		
		:param color: The :py:class:`Color` to use.
		:rtype: Nothing.
		"""
		self.color = color
		if self.brush and self.brush.doesUseSourceCaching():
			self.brush.cacheBrush(color)

	#Change only the alpha of the current color
	def setColorAlpha(self, fixed=None, proportional=None):
		"""
		Change the alpha of the current :py:class:`Color`.
		
		:param fixed: Set the absolute 0-1 value of the alpha.
		:param proportional: Set the relative value of the alpha (Es: If the current alpha is 0.8, a proportional value of 0.5 will set the final value to 0.4).
		:rtype: Nothing.
		"""
		if fixed!=None:
			self.color.set_alpha(fixed)
		elif proportional!=None:
			self.color.set_alpha(self.color.get_alpha()*proportional)

	#Gets the brush alpha
	def getColorAlpha(self):
		"""
		Retrieve the alpha of the current :py:class:`Color`.
		
		:rtype: A float 0-1 value of the current :py:class:`Color` alpha.
		"""
		return self.color.get_alpha()

	#Gets the brush size
	def getBrushSize(self):
		"""
		Retrieve the size of the current :py:class:`Brush`.
		
		:rtype: An integer value of the current :py:class:`Brush` size in pixels.
		"""
		return self.brush.brushSize

	#Swap between first and second color
	def swapColors(self):
		"""
		Swaps the current :py:class:`Color` with the secondary :py:class:`Color`.
		
		:rtype: Nothing.
		"""
		rgba = self.color.get_0_255()
		self.color = self.secondColor
		self.secondColor = Color(rgba, '0-255')

	#Setter for brush reference
	def setBrush(self, b, resize=0, proportional=None):
		"""
		Sets the size of the current :py:class:`Brush`.
		
		:param brush: The :py:class:`Brush` object to use as a brush.
		:param resize: An optional absolute value to resize the brush before using it.
		:param proportional: An optional relative float 0-1 value to resize the brush before using it.
		:rtype: Nothing.
		"""
		if proportional!=None:
			resize = int(self.brush.brushSize*0.5)
		b.resizeBrush(resize) #If resize=0 it reset to its default size
		self.brush = b
		if self.brush and self.brush.doesUseSourceCaching():
			self.brush.cacheBrush(self.color)

	#Setter for the mirror mode
	def setMirrorMode(self, mirror):
		"""
		Sets the mirror mode to use in the next operation.
		
		:param mirror: A string object with one of these values : '', 'h', 'v', 'hv'. "h" stands for horizontal mirroring, while "v" stands for vertical mirroring. "hv" sets both at the same time.
		:rtype: Nothing.
		"""
		assert (mirror=='' or mirror=='h' or mirror=='v' or mirror=='hv'or mirror=='vh'), 'setMirrorMode: wrong mirror mode, got '+str(mirror)+' expected one of ["","h","v","hv"]'
		
		#Round up all the coordinates and convert them to int		
		if mirror=='': 		mirror = 0
		elif mirror=='h': 	mirror = 1
		elif mirror=='v': 	mirror = 2
		elif mirror=='hv': 	mirror = 3
		elif mirror=='vh': 	mirror = 3
		self.mirrorMode = mirror
		
	#Render the final image
	def renderImage(self, output='', show=True):
		"""
		Renders the :py:class:`Image` and outputs the final PNG file.
		
		:param output: A string with the output file path, can be empty if you don't want to save the final image.
		:param show: A boolean telling the system to display the final image after the rendering is done.
		:rtype: Nothing.
		"""
		#Merge all the layers to apply blending modes
		resultLayer = self.image.mergeAllLayers()

		#Show and save the results
		img = PIL.Image.fromarray(resultLayer.data, 'RGBA')
		
		if show:
			img.show()

		if output!='':
			img.save(output, 'PNG')

	#Shortcut for image operations
	def newLayer(self, effect=''):
		"""
		Creates a new :py:class:`Layer` to the current :py:class:`Image`.
		
		:param effect: A string with the blend mode for that layer that will be used when during the rendering process. The accepted values are: :code:`'soft_light','lighten','screen','dodge','addition','darken','multiply','hard_light','difference','subtract','grain_extract','grain_merge','divide','overlay'`.
		:rtype: Nothing.
		"""
		self.image.newLayer(effect)

	#Shortcut for image operations
	def setActiveLayerEffect(self, effect):
		"""
		Changes the effect of the current active :py:class:`Layer`.
		
		:param output: A string with the one of the blend modes listed in :py:meth:`newLayer`.
		:rtype: Nothing.
		"""
		self.image.layers[self.image.activeLayer].effect = effect

	#Shortcut for image operations
	def duplicateActiveLayer(self):
		"""
		Duplicates the current active :py:class:`Layer`.
		
		:rtype: Nothing.
		"""
		self.image.duplicateActiveLayer()



















"""
The second most important module of the PaYnter project.
This module will manage all the data relative to the brushes you will create and use inside your scripts.

The :py:class:`Brush` is a quite complex class in the insides, but luckly the Paynter manages all the drawing.
So the final user of the library needs only to instantiate one (or more) brush and pass it to his :py:class:`Paynter`.
"""

#External modules
import PIL.Image
import numpy as N
from numba import jit, int32, float32, int64, float64, uint8, void

#PaYnter Modules
import paynter.config as config
from .utils import *
from .color import *

import time

######################################################################
# Brush class
######################################################################
resizeResample = PIL.Image.LANCZOS
rotateResample = PIL.Image.BICUBIC

#Image to brushtip function
def loadBrushTip(path, size, angle):
	#Open the image and make sure its RGB
	res = PIL.Image.open(config.ROOT+'/res/'+path).convert('RGB')
	
	#Resize it to the target size
	resScaled = res.resize((size, size), resample=resizeResample)
	
	#Invert and grab the value
	resScaled = PIL.ImageOps.invert(resScaled)
	resScaled = resScaled.rotate(angle, expand = 1)
	grayscaleValue = resScaled.split()[0]

	#Create the brushtip image
	bt = N.zeros((resScaled.width, resScaled.height, 4), dtype=N.float32)
	bt[:,:,0] = 1
	bt[:,:,1] = 1
	bt[:,:,2] = 1		
	bt[:,:,3] = N.divide(N.array(grayscaleValue), 255)

	#Return this 3D array full white and alpha 0-1 values
	return bt

#The brush class is the one that define how the current brush should behave
class Brush:
	"""
	The :py:class:`Brush` class is the one that defines how your Paynter should draw his lines.
	To create this class you can use the default constructor.
	   
	.. code-block:: python

		from paynter import *
		pixelBrush = Brush( "pixel.png", "", size = 100, spacing = 1)
	"""
	brushTip = 0
	brushMask = 0
	brushSize = 0
	multibrush = False
	spacing = 0
	fuzzyDabAngle = 0
	fuzzyDabSize = 0
	fuzzyDabHue = 0
	fuzzyDabSat = 0
	fuzzyDabVal = 0
	fuzzyDabMix = 0
	fuzzyDabScatter = 0
	originalSpacing = 0
	originalSize = 0
	realCurrentSize = 0
	coloredBrushSource = None 
	usesSourceCaching = False

	#Create the brush
	def __init__(self, tipImage, maskImage, size = 50, angle = 0, spacing = 1, 
				fuzzyDabAngle = 0, fuzzyDabSize = 0, fuzzyDabHue = 0, fuzzyDabSat = 0, fuzzyDabVal = 0, fuzzyDabMix = 0,
				fuzzyDabScatter = 0, usesSourceCaching = False):
		"""
		The creation of a :py:class:`Brush` can be as simple as the example above or pretty complex.
		The complexity is based on what effect you want to obtain from your Brush.
		There are a lot of parameters that you can define to customize your Brush the way you want to be.

		This is an example of a complex brush that emulates a watercolour brush:

		.. code-block:: python
		
			watercolor = Brush(["watercolor1.png","watercolor2.png","watercolor3.png","watercolor4.png","watercolor5.png"], # Use more than a single brush tip image to create more randomization
						"",				# No texture for this brush
						size = 440,			# Quite big brush
						angle = 0,			# No angle since we randomize it later each dab
						spacing = 0.5,			# Spacing to half the size of the brush
						fuzzyDabAngle = [0, 360],	# Randomize angle each dab
						fuzzyDabSize = [1, 2],		# Randomize the size of each dab between original size and double the size
						fuzzyDabHue = [-0.03, 0.03],	# Randomize slightly the hue of each dab
						fuzzyDabSat = [-0.2, 0.2],	# Randomize slightly the saturation of each dab
						fuzzyDabVal = [-0.1, 0.1],	# Randomize slightly the value of each dab
						fuzzyDabMix = [0.4, 0.5],	# Randomize slightly the mix of each dab
						fuzzyDabScatter = [0, 100])	# Make each dab go a bit of the trail to create more randomization


		:param tipImage: A string with the path of the image that you want to use as a tip. 
		                 If you want to load multiple brush tips and randomize between them each dab, you can place a list of strings instead.
		:param maskImage: A string with the path of the image that you want to use as a texture to apply to your brush.
		:param size: An integer with the size in pixel of the brush.
		:param angle: An integer with the angle (degrees) of the brush.
		:param spacing: A float with the spacing of the brush. Spacing is tells the system how far apart are each brush dab is. 
		                The value of this parameter is proportional to the brush size.
		:param fuzzyDabAngle: A list of two integers with the range between randomize the brush angle (degrees) of each dab. 
		:param fuzzyDabSize: A list of two numbers with the range between randomize the brush size (in a proportional way to the current brush size) of each dab.  
		:param fuzzyDabHue: A list of two float with the range between randomize the hue of each dab.
		:param fuzzyDabSat: A list of two float with the range between randomize the saturation of each dab.
		:param fuzzyDabVal: A list of two float with the range between randomize the value of each dab.
		:param fuzzyDabMix: A list of two float with the range between randomize the mix of each dab.
		:param fuzzyDabScatter: A list of two numbers with the range between randomize the deviation from the actual coordinates of each dab.
		:param usesSourceCaching: A boolean telling the system if he can cache the brush dab for much faster consecutives dab. 
		                          Not advised when the brush has fuzzy dab parameters.
		"""


		#Downsample the size-related parameters
		self.originalSize = size
		self.realCurrentSize = size
		size = int(size/config.DOWNSAMPLING)
		if fuzzyDabScatter!=0:
			fuzzyDabScatter[0] = int(fuzzyDabScatter[0]/config.DOWNSAMPLING)
			fuzzyDabScatter[1] = int(fuzzyDabScatter[1]/config.DOWNSAMPLING)

		#Set the brushTip
		if isinstance(tipImage, list):
			#Multibrush
			self.brushTip = []
			for image in tipImage:
				bt = loadBrushTip(image, size, angle)
				self.brushTip.append(bt)
			self.multibrush = True
			self.brushSize = self.brushTip[0].shape[0]
		else:
			#NormalBrush
			self.brushTip = loadBrushTip(tipImage, size, angle)
			self.brushSize = self.brushTip.shape[0]
		
		#Set the perameters
		self.spacing = size*spacing
		self.originalSpacing = spacing
		self.fuzzyDabAngle = fuzzyDabAngle
		self.fuzzyDabSize = fuzzyDabSize
		self.fuzzyDabHue = fuzzyDabHue
		self.fuzzyDabSat = fuzzyDabSat
		self.fuzzyDabVal = fuzzyDabVal
		self.fuzzyDabMix = fuzzyDabMix
		self.fuzzyDabScatter = fuzzyDabScatter
		self.usesSourceCaching = usesSourceCaching

		#Set the brush mask
		if maskImage!="":
			res = PIL.Image.open(config.ROOT+'/res/'+maskImage)
			while res.width>config.CANVAS_SIZE:
				res = res.resize((int(res.width/2),int(res.width/2)), resample=resizeResample)
			alpha = res.split()[0] #Take only the first channel since is black and white
			bm = N.zeros((config.CANVAS_SIZE, config.CANVAS_SIZE), dtype=N.float32)
			for j in range(0, config.CANVAS_SIZE, res.width):
				for i in range(0, config.CANVAS_SIZE, res.width):
					bm[i:i+res.width, j:j+res.width] = N.divide(N.array(alpha), 255)
			self.brushMask = 1-bm
		
		else:
			#Add default mask
			bm = N.zeros((config.CANVAS_SIZE, config.CANVAS_SIZE), dtype=N.float32)
			bm[:,:] = 0
			self.brushMask = 1-bm

	#Get the usesSourceCaching
	def doesUseSourceCaching(self):
		return self.usesSourceCaching

	#Cache the brush to save processing time
	def cacheBrush(self, color):
		self.coloredBrushSource = self.brushTip[:,:] * color.get_0_1()

	#Resize brush
	def resizeBrush(self, newSize):
		#Check if I want to reset back to original
		if newSize==0:
			newSize = self.originalSize

		#Don't do useless calculations
		if self.realCurrentSize == newSize:
			return

		#Downsample the size-related parameters
		print('resizing from :'+str(self.realCurrentSize)+' to:'+str(newSize))
		self.realCurrentSize = newSize
		newSize = max(int(newSize/config.DOWNSAMPLING), 1)

		#Set the brushTip
		if self.multibrush:
			#Multibrush
			for i in range(0,len(self.brushTip)):
				btImg = PIL.Image.fromarray((self.brushTip[i]*255).astype(N.uint8), 'RGBA')
				btImgScaled = btImg.resize((newSize, newSize), resample=resizeResample)
				btArray = N.divide(N.array(btImgScaled).astype(N.float32), 255)
				self.brushTip[i] = btArray
			self.brushSize = self.brushTip[0].shape[0]
		else:
			#NormalBrush
			btImg = PIL.Image.fromarray((self.brushTip*255).astype(N.uint8), 'RGBA')
			btImgScaled = btImg.resize((newSize, newSize), resample=resizeResample)
			btArray = N.divide(N.array(btImgScaled).astype(N.float32), 255)
			self.brushTip = btArray
			self.brushSize = self.brushTip.shape[0]
		
		#Set the perameters
		self.spacing = newSize*self.originalSpacing

	#Prepare a dab before the applyDab function, triggered only on brushes that do NOT use sourceCaching since each dab has different source on those
	def prepareDab(self, color, secondColor):
		#Get the brush image image 
		brushSource = self.brushTip if not self.multibrush else self.brushTip[randInt(0,len(self.brushTip)-1)]

		#Apply transformations
		if self.fuzzyDabAngle!=0 or self.fuzzyDabSize!=0:
			#Convert the brush to an image to ease out the rotation and resizing process
			img = PIL.Image.fromarray((brushSource*255).astype(N.uint8), 'RGBA')
			
			#Apply fuzzy scale
			if self.fuzzyDabAngle!=0:
				img = img.rotate(fuzzy(self.fuzzyDabAngle), expand=1, resample=PIL.Image.BILINEAR)
		
			#Apply fuzzy scale
			if self.fuzzyDabSize!=0:
				fuz = fuzzy(self.fuzzyDabSize)
				img = img.resize((int(img.width*fuz), int(img.height*fuz)), resample=PIL.Image.BILINEAR)
			
			#Reconvert brushSource to an array
			brushSource = N.array(img, dtype=N.float32)*0.0039

		if self.fuzzyDabHue!=0 or self.fuzzyDabSat!=0 or self.fuzzyDabVal!=0 or self.fuzzyDabMix!=0:
			#Apply fuzzy color transformations
			tmpColor = color.copy()
			
			#Mix colors
			if self.fuzzyDabMix!=0:
				fuz = fuzzy(self.fuzzyDabMix)
				tmpColor.r = min(1, (tmpColor.r*(1-fuz) + secondColor.r*fuz))
				tmpColor.g = min(1, (tmpColor.g*(1-fuz) + secondColor.g*fuz))
				tmpColor.b = min(1, (tmpColor.b*(1-fuz) + secondColor.b*fuz))
				
			#Fuzzy Hue
			if self.fuzzyDabHue!=0:
				tmpColor.tweak_Hue(fuzzy(self.fuzzyDabHue))

			#Fuzzy Saturation
			if self.fuzzyDabSat!=0:
				tmpColor.tweak_Sat(fuzzy(self.fuzzyDabSat))

			#Fuzzy Value
			if self.fuzzyDabVal!=0:
				tmpColor.tweak_Val(fuzzy(self.fuzzyDabVal))

			color = tmpColor

		#Color the brush
		return brushSource[:,:] * color.get_0_1()

	#Make a single dab on the canvas
	def makeDab(self, layer, x, y, color, secondColor, mirror=0):
		#Prepare the dab only if i don't have source caching
		if not self.usesSourceCaching:
			#Prepare the dab with all the fuzzy parameters 
			self.coloredBrushSource = self.prepareDab(color, secondColor)

		#Apply scattering if any
		if self.fuzzyDabScatter != 0:
			randomAngle = random.uniform(0,6.28)
			randomLength = fuzzy(self.fuzzyDabScatter)
			x += N.cos(randomAngle)*randomLength
			y += N.sin(randomAngle)*randomLength

		#Use the Jitted version to optimized calculus
		applyMirroredDab_jit(mirror, layer.data, int(x-self.brushSize*0.5), int(y-self.brushSize*0.5), self.coloredBrushSource.copy(), config.CANVAS_SIZE, self.brushMask)
		

#Jitted version of applyDab
@jit(void(uint8[:,:,:], int64, int64, float32[:,:,:], int64, float32[:,:]), nopython=True, cache=True)
def applyDab_jit(layerData, x, y, source, canvSize, brushMask):
	#Get the final dab size
	dabSizeX, dabSizeY = source.shape[:2]
	
	#Adjust coordinates and make sure we are inside (at least partially) the canvas
	adj_x1, adj_y1 = int(max(0, min(canvSize, x))), 			int(max(0, min(canvSize, y)))
	adj_x2, adj_y2 = int(max(0, min(canvSize, x+dabSizeX))), 	int(max(0, min(canvSize, y+dabSizeY)))
	if adj_x1==adj_x2 or adj_y1==adj_y2:return
	
	#Get the slice and uniform to [0-1]
	destination = N.divide(N.copy(layerData[adj_y1:adj_y2, adj_x1:adj_x2].astype(N.float32)), 255)
	
	#Calculate the correct range to make sure it works even on canvas border 
	bx1 = 0 if (x>=0) else dabSizeX-adj_x2
	bx2 = bx1+dabSizeX if (x+dabSizeX<canvSize) else canvSize - adj_x1
	by1 = 0 if (y>=0) else dabSizeY-adj_y2
	by2 = by1+dabSizeY if (y+dabSizeY<canvSize) else canvSize - adj_y1
	
	#Color the brush, slice it if is on the canvas border, and apply the brush texture on it
	source = source[by1:by2, bx1:bx2,:]
	source[:,:,3] *= brushMask[adj_y1:adj_y2, adj_x1:adj_x2]
	
	#Apply source image over destination using the SRC alpha ADD DEST inverse_alpha blending method
	inverseSource = N.subtract(1, source[:,:,3])
	normalSource = source[:,:,3]
	layerData[adj_y1:adj_y2, adj_x1:adj_x2, 0] = (((destination[:,:,0] * inverseSource) + (source[:,:,0] * normalSource))*255).astype(N.uint8)
	layerData[adj_y1:adj_y2, adj_x1:adj_x2, 1] = (((destination[:,:,1] * inverseSource) + (source[:,:,1] * normalSource))*255).astype(N.uint8)
	layerData[adj_y1:adj_y2, adj_x1:adj_x2, 2] = (((destination[:,:,2] * inverseSource) + (source[:,:,2] * normalSource))*255).astype(N.uint8)
	layerData[adj_y1:adj_y2, adj_x1:adj_x2, 3] = ((destination[:,:,3] + (1 - destination[:,:,3]) * normalSource)*255).astype(N.uint8)


#Jitted version of applyMirroredDab
@jit(void(int64, uint8[:,:,:], int64, int64, float32[:,:,:], int64, float32[:,:]), nopython=True, cache=True)
def applyMirroredDab_jit(mirror, layerData, x, y, source, canvSize, brushMask):
	#Apply the first normal dab
	applyDab_jit(layerData, x, y, source, canvSize, brushMask)

	#Mirrored horizontally
	if mirror==1 or mirror==3:
		brushSourceFlipped = N.copy(source)[:,::-1]
		applyDab_jit(layerData, layerData.shape[0]-x-brushSourceFlipped.shape[0], y, brushSourceFlipped, canvSize, brushMask)

	#Mirrored vertically
	if mirror==2 or mirror==3:
		brushSourceFlipped = N.copy(source)[::-1]
		applyDab_jit(layerData, x, layerData.shape[1]-y-source.shape[1], source, canvSize, brushMask)
	
	#Double mirrored
	if mirror==3:
		brushSourceFlipped = N.copy(source)[:,::-1]
		brushSourceFlippedFlipped = N.copy(brushSourceFlipped)[::-1]
		applyDab_jit(layerData, layerData.shape[0]-x-brushSourceFlippedFlipped.shape[0], layerData.shape[1]-y-brushSourceFlippedFlipped.shape[1], brushSourceFlippedFlipped, canvSize, brushMask)












#Python base libs
import sys

#Here it goes all the variables/constant that are shared between modules
CANVAS_SIZE = 0
DEBUG = False
DOWNSAMPLING = 1
REAL_CANVAS_SIZE = 0
ROOT = sys.path[0]


AVGTIME = []


"""
Another key class inside the Paynter library is the Image class.
This module will take care of storing all the layers data of the current image you are creating.
It will also manage to merge all the layer during the final rendering process.

This class will normally be 100% managed through the :py:class`Paynter` class, so you should never instantiate it manually.
"""
#External modules
import PIL.Image 
import numpy as N

#PaYnter Modules
from .layer import Layer
from .blendModes import *

import time

######################################################################
# Image Functions
######################################################################
#The image class is a container for all the layers of the image
class Image:
	"""
	The :py:class:`Image` class is structured as an array of :py:class:`Layer`.

	An Image class is created when you create a :py:class:`Paynter`, so you can access this class as follows:

	.. code-block:: python

		from paynter import *
		
		paynter = Paynter()
		image = paynter.image
	"""
	layers = []
	activeLayer = 0

	#Init image
	def __init__(self):
		#Init by adding a new layer and selecting that as current layer
		self.layers.append(Layer())
		self.activeLayer = len(self.layers)-1

	#Return the current active layer
	def getActiveLayer(self):
		"""
		Returns the currently active :py:class:`Layer`.
		
		:rtype: A :py:class:`Layer` object.
		"""
		return self.layers[self.activeLayer]

	#Create a new layer and select it
	def newLayer(self, effect=''):
		"""
		Creates a new :py:class:`Layer` and set that as the active.
		
		:param effect: A string with the blend mode for that layer that will be used when during the rendering process. The accepted values are: :code:`'soft_light','lighten','screen','dodge','addition','darken','multiply','hard_light','difference','subtract','grain_extract','grain_merge','divide','overlay'`.
		:rtype: Nothing.
		"""
		self.layers.append(Layer(effect = effect))
		self.activeLayer = len(self.layers)-1

	#Duplicate the current layer
	def duplicateActiveLayer(self):
		"""
		Duplicates the current active :py:class:`Layer`.
		
		:rtype: Nothing.
		"""
		activeLayer = self.layers[self.activeLayer]
		newLayer = Layer(data=activeLayer.data, effect=activeLayer.effect)
		self.layers.append(newLayer)
		self.activeLayer = len(self.layers)-1

	#Merge all the layers together to render the final image
	def mergeAllLayers(self):
		"""
		Merge all the layers together.

		:rtype: The result :py:class:`Layer` object. 
		"""
		start = time.time()
		while(len(self.layers)>1):
			self.mergeBottomLayers()
		print('merge time:'+str(time.time()-start))
		return self.layers[0]

	#Merge the bottom layer with the layer above that
	def mergeBottomLayers(self):
		#Debug show the two layer being merged
		print('merging layers with:'+str(self.layers[1].effect))
		#Normal paste on top
		if self.layers[1].effect=='':
			baseImage = PIL.Image.fromarray(self.layers[0].data, 'RGBA')
			overImage = PIL.Image.fromarray(self.layers[1].data, 'RGBA')
			baseImage = PIL.Image.alpha_composite(baseImage, overImage)
			newImage = N.array(baseImage)
		#Apply blend mode
		else:
			baseImage = self.layers[0].data.astype(N.float32)
			overImage = self.layers[1].data.astype(N.float32)
			newImage = mergeImagesWithBlendMode(baseImage, overImage, self.layers[1].effect).astype(N.uint8)
		#Remove one layer and replace the last one
		del self.layers[0]			
		self.layers[0] = Layer(data = newImage)

	


#Import all the paynter modules
name = "paynter"


from .brush import *
from .paynter import *
from .layer import *
from .image import *
from .color import *
from .blendModes import *
import paynter.config as config



#External modules
import numpy as N 
from numba import jit, float32, int64

#Python base libs
import random
import math
import colorsys

######################################################################
# Useful Functions
######################################################################
#Jitted functions for trigonometry
@jit(float32(float32, float32), nopython=True, cache=True)
def lendir_x(l, d):
	return l*N.cos(d)

#Jitted functions for trigonometry
@jit(float32(float32, float32), nopython=True, cache=True)
def lendir_y(l, d):
	return l*N.sin(d)

#Clamp shortcut
def clamp(x, mi, ma):
	return max(mi, min(ma, x))

#Random shortcut
def randInt(a,b):
	if int(a)==int(b):
		return int(a)
	if a>b:
		tmp = b 
		b = a 
		a = b
	return random.randint(int(a),int(b)) #inclusive

#Random shortcut
def randFloat(a,b):
	return random.uniform(a,b)

#Degree cosine shortcut
def dcos(deg):
	return math.cos(math.radians(deg))

#Degree sine shortcut
def dsin(deg):
	return math.sin(math.radians(deg))

#Get fuzzy
def fuzzy(fuzzyRange):
	return random.uniform(fuzzyRange[0], fuzzyRange[1])

#Rotate a list of points around a center for an angle
def rotateMatrix(pointList, cx, cy, angle):	
	rotatedPoints = []
	#For each point in the list
	for point in pointList:
		#Grab the coords and get dir and len
		oldX = point[0]
		oldY = point[1]
		direction = math.degrees(math.atan2(cy-oldY, cx-oldX))
		length = math.sqrt((cx - oldX)**2 + (cy - oldY)**2)

		#Rotate them and insert in the return list
		newX = cx+length*dcos(direction+angle)
		newY = cy+length*dsin(direction+angle)
		rotatedPoints.append([newX, newY])
	return rotatedPoints

#Interpolate two points (pos=0 -> returns x1,y1; pos=1 -> return x2,y2)
def linearInterpolationPoint(x1,y1,x2,y2, pos):
    xx = x1+pos*(x2-x1)
    yy = y1+pos*(y2-y1)
    return (xx,yy)



"""
The Color class is another fundamental class in the Paynter library.
This module will mange the creation, modification, and storing of colors and palettes.

The color class is mainly used internally by the :py:class:`Paynter` class, but the user will still have to create the palette and sets the active colors manually through Paynter.setColor(color).
"""

#PaYnter Modules
from .utils import *
from numba import jitclass, float32, float64

import sys







@jit([float32[:](float32, float32, float32)], nopython=True)
def rgb_to_hsv(r, g, b):
    maxc = max(r, g, b)
    minc = min(r, g, b)
    v = maxc
    if minc == maxc:
        return N.array((0.0, 0.0, v), dtype=N.float32)
    s = (maxc-minc) / maxc
    rc = (maxc-r) / (maxc-minc)
    gc = (maxc-g) / (maxc-minc)
    bc = (maxc-b) / (maxc-minc)
    if r == maxc:
        h = bc-gc
    elif g == maxc:
        h = 2.0+rc-bc
    else:
        h = 4.0+gc-rc
    h = (h/6.0) % 1.0
    return N.array((h, s, v), dtype=N.float32)

@jit(float32[:](float32, float32, float32), nopython=True)
def hsv_to_rgb(h, s, v):
    if s == 0.0:
        return N.array((v, v, v), dtype=N.float32)
    i = int(h*6.0) # XXX assume int() truncates!
    f = (h*6.0) - i
    p = v*(1.0 - s)
    q = v*(1.0 - s*f)
    t = v*(1.0 - s*(1.0-f))
    i = i%6
    if i == 0:
        return N.array((v, t, p), dtype=N.float32)
    if i == 1:
        return N.array((q, v, p), dtype=N.float32)
    if i == 2:
        return N.array((p, v, t), dtype=N.float32)
    if i == 3:
        return N.array((p, q, v), dtype=N.float32)
    if i == 4:
        return N.array((t, p, v), dtype=N.float32)
    if i == 5:
        return N.array((v, p, q), dtype=N.float32)
    return N.array((0,0,0), dtype=N.float32)




######################################################################
# Color Management functions
######################################################################
#Get 3 colors with a triad pattern
def getColors_Triad(hue=None, sat = 1, val = 1, spread = 60):
	"""
	Create a palette with one main color and two opposite color evenly spread apart from the main one. 

	:param hue: A 0-1 float with the starting hue value.
	:param sat: A 0-1 float with the palette saturation.
	:param val: A 0-1 float with the palette value.
	:param val: An int with the spread in degrees from the opposite color.
	:rtype: A list of :py:class:`Color` objects.
	"""
	palette = list()
	if hue==None:
		leadHue = randFloat(0, 1)
	else:
		leadHue = hue
	palette.append(Color(0,0,0,1).set_HSV(leadHue, sat, val))
	palette.append(Color(0,0,0,1).set_HSV((leadHue + 0.5 + spread/360) % 1, sat, val))
	palette.append(Color(0,0,0,1).set_HSV((leadHue + 0.5 - spread/360) % 1, sat, val))	
	return palette 



#Colors are always internally stored as 0-1 floats
spec = [
    ('r', float32),
    ('g', float32),
    ('b', float32),
    ('a', float32)
]

@jitclass(spec)
class Color:
	"""
	The :py:class:`Color` class has 4 0-1 floats, one for each RGBA channel.

	A Color class is created from palette functions or directly with their constructor.

	.. code-block:: python

		from paynter import *
		
		#Get a list of colors
		palette = getColors_Triad(spread = 20)

		#You can create with 0-1 floats..
		otherColor = Color(1, 0.5, 0, 1)

		#.. or with 0-255 ints
		sameColor = Color(255, 128, 0, 255)
	"""
	def __init__(self, r, g, b, a):
		if max((r,g,b,a))>1:
			r /= 255
			b /= 255
			g /= 255
			a /= 255

		self.r = r
		self.g = g
		self.b = b
		self.a = a

	#Return 0-255 RGBA
	def get_0_255(self):
		"""
		Gets the RGBA 0-255 rappresentation of this color.

		:rtype: A list of 4 ints.
		"""
		return [self.r*255, self.g*255, self.b*255, self.a*255]

	#Return 0-1 RGBA
	def get_0_1(self):
		"""
		Gets the RGBA 0-1 rappresentation of this color.

		:rtype: A list of 4 floats.
		"""
		return N.array((self.r, self.g, self.b, self.a), dtype=N.float32)

	#Returns 0-1 HSV
	def get_HSV(self):
		"""
		Gets the HSV 0-1 rappresentation of this color.
		This does ignore the alpha value.

		:rtype: A list of 3 floats.
		"""
		return rgb_to_hsv(self.r, self.g, self.b)

	#Overwrite RGB values with this HSV
	def set_HSV(self, h, s, v):
		"""
		Overwrite the current color with this set of HSV values.
		This keeps the current alpha value.
		
		:param h: A 0-1 float with the Hue.
		:param s: A 0-1 float with the Saturation.
		:param v: A 0-1 float with the Value.
		:rtype: The new :py:class:`Color` object.
		"""
		rgb = hsv_to_rgb(h,s,v)
		self.r = rgb[0]
		self.g = rgb[1]
		self.b = rgb[2]
		return self

	#Sets a new alpha value
	def set_alpha(self, newAlpha):
		"""
		Overwrite the current alpha value.
		
		:param newAlpha: A 0-1 float with the desired alpha.
		:rtype: The :py:class:`Color` object.
		"""
		self.a = newAlpha
		return self

	#Gets the alpha value
	def get_alpha(self):
		"""
		Gets the current alpha value.

		:rtype: A 0-1 float.
		"""
		return self.a

	#Tweak the hue
	def tweak_Hue(self, ammount):
		"""
		Change the current hue value by a certain ammount.

		:param ammount: A 0-1 float with the ammount to sum to the current hue.
		:rtype: The :py:class:`Color` object.
		"""
		hsv = self.get_HSV()
		hsv[0] = (hsv[0] + ammount) % 1
		self.set_HSV(hsv[0], hsv[1], hsv[2])
		return self

	#Tweak the saturation
	def tweak_Sat(self, ammount):
		"""
		Change the current saturation value by a certain ammount.

		:param ammount: A 0-1 float with the ammount to sum to the current saturation.
		:rtype: The :py:class:`Color` object.
		"""
		hsv = self.get_HSV()
		hsv[1] = min(max(hsv[1] + ammount, 0), 1)
		self.set_HSV(hsv[0], hsv[1], hsv[2])
		return self
		
	#Tweak the value
	def tweak_Val(self, ammount):
		"""
		Change the current value value by a certain ammount.

		:param ammount: A 0-1 float with the ammount to sum to the current value.
		:rtype: The :py:class:`Color` object.
		"""
		hsv = self.get_HSV()
		hsv[2] = min(max(hsv[2] + ammount, 0), 1)
		self.set_HSV(hsv[0], hsv[1], hsv[2])
		return self

	def copy(self):
		"""
		Creates a copy of this Color.

		:rtype: The new :py:class:`Color` object.
		"""
		return Color(self.r, self.g, self.b, self.a)
		























import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()



setuptools.setup(
    name="paynter",
    version="0.1.4",
    author="Stefano Bertoli",
    author_email="elkiwydev@gmail.com",
    description="PaYnter is a Python Library that let you procedurally generate images with handy features that emulates what you can find in any image editing software like Photoshop, GIMP, Krita, and similars.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/elkiwy/paynter",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)


"""
This is the class that compose the :py:class:`Image` objects.
This module will take care of storing data as 3D arrays of uint8.
The array dimension are width, height, and RGBA channels of that layer.

This class will be 100% managed through the :py:class`Paynter` class, so you should never instantiate it manually.
"""

#External modules
import numpy as N
import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont

#PaYnter Modules
import paynter.config as config
from .color import *

######################################################################
# Layer Functions
######################################################################
#The layer class the the one holding all the data and effecct to show how to add colors when merging layers
class Layer:
	"""
	The :py:class:`Layer` class contains a 3D array of N.uint8 and a string with the blend mode of the layer.

	An Image starts with one layer inside, but you can create more of them as follows:

	.. code-block:: python

		from paynter import *

		#Inside the paynter there is already an Image with a blank Layer.
		paynter = Paynter()
		
		#Create a blank new layer
		payner.newLayer()

		#Create a new layer duplicating the current one
		payner.duplicateActiveLayer()

		#Gets the current active layer		
		layer = paynter.image.getActiveLayer()
	"""
	effect = ''
	data = 0

	#Layer constructor
	# Layers data structure are always 0-255 ints
	def __init__(self, data = None, color = None, effect = ''):
		if type(data) is not N.ndarray:
			self.data = N.zeros((config.CANVAS_SIZE, config.CANVAS_SIZE, 4), dtype=N.uint8)
			if color==None:
				color = Color(0,0,0,0)
			colorRGBA = color.get_0_255()
			self.data[:,:,0] = colorRGBA[0]
			self.data[:,:,1] = colorRGBA[1]
			self.data[:,:,2] = colorRGBA[2]
			self.data[:,:,3] = colorRGBA[3]
		else:
			self.data = data
		self.effect = effect

	#Show layer as a separate image with the option to write on it 
	def showLayer(self, title='', debugText=''):
		"""
		Shows the single layer.
		
		:param title: A string with the title of the window where to render the image.
		:param debugText: A string with some text to render over the image.
		:rtype: Nothing.
		"""
		img = PIL.Image.fromarray(self.data, 'RGBA')
		if debugText!='':
			draw = PIL.ImageDraw.Draw(img)
			font = PIL.ImageFont.truetype("DejaVuSansMono.ttf", 24)
			draw.text((0, 0),debugText,(255,255,255),font=font)
		img.show(title=title)



'''
This file is a custom and more compact version of 'blend_modes' made by Florian Roscheck.
You can find more about it here at https://pythonhosted.org/blend_modes/
'''
import numpy as np

#Works on 0-255 float arrays and outputs a 0-255 float array
def mergeImagesWithBlendMode(img_in, img_layer, blendMode, opacity=1):
    #Take the 0-255 input and convert to 0-1 range
    img_in /= 255
    img_layer /= 255.0
    img_out = 0

    #Compose alpha 
    comp_alpha = np.minimum(img_in[:, :, 3], img_layer[:, :, 3])*opacity
    new_alpha = img_in[:, :, 3] + (1.0 - img_in[:, :, 3])*comp_alpha
    np.seterr(divide='ignore', invalid='ignore')
    ratio = comp_alpha/new_alpha
    ratio[ratio == np.NAN] = 0.0
    comp = 0

    #Soft light
    if blendMode=='soft_light':
        comp = (1.0 - img_in[:, :, :3]) * img_in[:, :, :3] * img_layer[:, :, :3] \
                   + img_in[:, :, :3] * (1.0 - (1.0-img_in[:, :, :3])*(1.0-img_layer[:, :, :3]))
    
    #Lighten only    
    elif blendMode=='lighten':
        comp = np.maximum(img_in[:, :, :3], img_layer[:, :, :3])

    #Screen
    elif blendMode=='screen':
        comp = 1.0 - (1.0 - img_in[:, :, :3]) * (1.0 - img_layer[:, :, :3])

    #Dodge
    elif blendMode=='dodge':
        comp = np.minimum(img_in[:, :, :3]/(1.0 - img_layer[:, :, :3]), 1.0)

    #Addition
    elif blendMode=='addition':
        comp = img_in[:, :, :3] + img_layer[:, :, :3]

    #Darken only
    elif blendMode=='darken':
        comp = np.minimum(img_in[:, :, :3], img_layer[:, :, :3])

    #Multiply
    elif blendMode=='multiply':
        comp = np.clip(img_layer[:, :, :3] * img_in[:, :, :3], 0.0, 1.0)

    #Hard light
    elif blendMode=='hard_light':
         comp = np.greater(img_layer[:, :, :3], 0.5)*np.minimum(1.0-((1.0 - img_in[:, :, :3]) * (1.0 - (img_layer[:, :, :3] - 0.5) * 2.0)), 1.0) \
           + np.logical_not(np.greater(img_layer[:, :, :3], 0.5))*np.minimum(img_in[:, :, :3] * (img_layer[:, :, :3] * 2.0), 1.0)

    #Difference
    elif blendMode=='difference':
        comp = img_in[:, :, :3] - img_layer[:, :, :3]
        comp[comp < 0.0] *= -1.0

    #Subtract    
    elif blendMode=='subtract':
        comp = img_in[:, :, :3] - img_layer[:, :, :3]

    #Grain extract
    elif blendMode=='grain_extract':
        comp = np.clip(img_in[:, :, :3] - img_layer[:, :, :3] + 0.5, 0.0, 1.0)

    #Grain merge
    elif blendMode=='grain_merge':
        comp = np.clip(img_in[:, :, :3] + img_layer[:, :, :3] - 0.5, 0.0, 1.0)

    #Divide
    elif blendMode=='divide':
        comp = np.minimum((256.0 / 255.0 * img_in[:, :, :3]) / (1.0 / 255.0 + img_layer[:, :, :3]), 1.0)

    #Overlay
    elif blendMode=='overlay':
        comp = img_in[:,:,:3] * (img_in[:,:,:3] + (2 * img_layer[:,:,:3]) * (1 - img_in[:,:,:3]))

    else:
        print('ERROR: wrong blendMode:'+str(blendMode))

    #Compose img_out
    ratio_rs = np.reshape(np.repeat(ratio, 3), [comp.shape[0], comp.shape[1], comp.shape[2]])
    img_out = np.clip(comp * ratio_rs + img_in[:, :, :3] * (1.0 - ratio_rs), 0.0, 1.0)
    img_out = np.nan_to_num(np.dstack((img_out, img_in[:, :, 3])))  # add alpha channel and replace nans
    return img_out*255.0





